#agents.py
import random
from model.model import *
import order.order
from model.model_statistics import *
from agents.strategic import StrategicAgent
from agents.clearing_price import ClearingPriceAgent
from agents.fixed_markup import FixedMarkupAgent
from agents.random_markup import RandomMarkupAgent

def buy_unit_portfolio(agent, model):
    # Wealth is enough to buy a unit portfolio. If the agent thinks that buying a unit portfolio is cheaper than 
    # buying each contract on the market, the agent will buy a unit portfolio.
    while((agent.wealth >= 1) & ((get_lowest_ask(model.ask_queue_alpha).price + get_lowest_ask(model.ask_queue_beta).price >= 1))):
        agent.contracts_alpha += 1
        agent.contracts_beta += 1
        agent.wealth -= 1
        model.unit_portfolios += 1
    return

def sell_unit_portfolio(agent, model):
    # Agent owns a contract of each (unit portfolio). If the agent thinks that selling a unit portfolio earns more
    # than selling each contract on the market, the agent will sell a unit portfolio.
    while(((agent.contracts_alpha >= 1) & (agent.contracts_beta >= 1)) & (get_highest_bid(model.bid_queue_alpha).price + get_highest_bid(model.bid_queue_beta).price<=1)):
        agent.contracts_alpha -= 1
        agent.contracts_beta -= 1
        agent.wealth += 1
        model.unit_portfolios -= 1
    return

# A function to determine the expected value of buying a contract
def utility(price, valuation, model, wealth):
    if((wealth >= price) & (valuation>price) & (price>0)):
        return (valuation * (1 - price))/price
    return 0

# The buyers "pays" the seller and the trade is added to the trades data (which is used as output)
def transaction(buyer, seller, price, trades):
    buyer.wealth -= price
    seller.wealth += price
    trades.append(price)

# The buyer gets an alpha contract from the seller
def alpha_transaction(buyer, seller, price, model):
    transaction(buyer,seller,price, model.trades_alpha)
    buyer.contracts_alpha += 1
    seller.contracts_alpha -= 1
    return

# The buyer gets a beta contract from the seller
def beta_transaction(buyer, seller, price, model):
    transaction(buyer,seller,price, model.trades_beta)
    buyer.contracts_beta += 1
    seller.contracts_beta -= 1
    return

# A new ask is added to the queue
def append_ask_queue(queue, order):
    if len(queue)>=200: # The queue is limited to 200, only the lowest asks remain
        if queue[-1].price > order.price:
            del queue[-1] # delete the highest from the queue
        else:
            return
    queue.append(order) # Always append, whether len>=200 or not, because we delete one when higher than 200
    queue.sort(key=lambda order: order.price)   

# A new bid is added to the queue
def append_bid_queue(queue, order):
    if len(queue)>=200: # The queue is limited to 200, only the highest bids remain
        if queue[0].price < order.price:
            del queue[-1] # delete the lowest from the queue
        else:
            return
    queue.append(order) # Always append, whether len>=200 or not, because we delete one when higher than 200
    queue.sort(key=lambda order: order.price)   

# Places a new ask for an alpha contract
def place_new_ask_alpha(agent, price, model):
    highest_bid = get_highest_bid(model.bid_queue_alpha)
    # There is a bid that matches the ask, the market clears
    if ((highest_bid.price >= price) & (len(model.bid_queue_alpha)>0)):
        alpha_transaction(highest_bid.agent, agent, price, model)
        model.bid_queue_alpha.pop()
    else:
        append_ask_queue(model.ask_queue_alpha, Order(price,agent)) # No clearing bid, we add the ask to the queue

# Places a new ask for a beta contract
def place_new_ask_beta(agent, price, model):
    highest_bid = get_highest_bid(model.bid_queue_beta)
    # There is a bid that matches the ask, the market clears
    if (((highest_bid.price >= price) & (len(model.bid_queue_beta)>0))):
        beta_transaction(highest_bid.agent, agent, price, model)
        model.bid_queue_beta.pop()
    else:
        append_ask_queue(model.ask_queue_beta, Order(price,agent)) # No clearing bid, we add the ask to the queue
 
# Places a new bid for an alpha contract
def place_new_bid_alpha(agent, price, model):
    lowest_ask = get_lowest_ask(model.ask_queue_alpha)
    # There is a ask that matches the bid, the market clears
    if(((lowest_ask.price <= price) & (len(model.ask_queue_alpha)>0))):
        alpha_transaction(agent, lowest_ask.agent, price, model)
        model.ask_queue_alpha.pop()
    else:
        append_bid_queue(model.bid_queue_alpha, Order(price,agent)) # No clearing ask, we add the bid to the queue

# Places a new bid for a beta contract
def place_new_bid_beta(agent, price, model):
    lowest_ask = get_lowest_ask(model.ask_queue_beta)
    # There is a ask that matches the bid, the market clears
    if (((lowest_ask.price <= price) & (len(model.ask_queue_beta)>0))):
        beta_transaction(agent, lowest_ask.agent, price, model)
        model.ask_queue_beta.pop()
    else:
        append_bid_queue(model.bid_queue_beta, Order(price,agent)) # No clearing ask, we add the bid to the queue
 
# Places new asks       
def place_asks(agent,model):
    temp_contracts_alpha = agent.contracts_alpha # All asks needs to be fullfilable at any time (number of asks <= contracts)
    temp_contracts_beta = agent.contracts_beta
    remove_from_ask_queues(model, agent.unique_id) # Remove all old asks from the queues
    while(temp_contracts_alpha>0):  # Place new asks for the number of alpha contracts
        place_new_ask_alpha(agent, agent.ask_price_alpha(model), model)
        temp_contracts_alpha -= 1
    while(temp_contracts_beta>0): # Place new asks for the number of beta contracts
        place_new_ask_beta(agent, agent.ask_price_beta(model), model)
        temp_contracts_beta -= 1

# Place a new bid for alpha
def bid_alpha(agent, price_alpha, model, temp_wealth):
    place_new_bid_alpha(agent, price_alpha, model)
    return temp_wealth - price_alpha

# Place a new bid for beta
def bid_beta(agent, price_beta, model, temp_wealth):
    place_new_bid_beta(agent, price_beta, model)
    return temp_wealth - price_beta

# If buying an alpha contract and a beta contract have the same expected value,
# use a random function to determine which one to buy
def decide_equal_utility(agent, price_alpha, price_beta, model, temp_wealth):
    rand = random.randint(0,1)
    if(rand == 0):
        return bid_alpha(agent, price_alpha, model, temp_wealth)
    else:
        return bid_beta(agent, price_beta, model, temp_wealth)

# Determine which contract has the highest utility and make a bid for that contract
def decide_contract(agent, model, price_alpha, price_beta, temp_wealth):
    u_alpha = utility(price_alpha, agent.belief, model, agent.wealth)
    u_beta = utility(price_beta, 1-agent.belief, model, agent.wealth)
    if((u_alpha > u_beta) & (u_alpha>0) & (price_alpha>0) & (price_alpha<=temp_wealth)):
        return bid_alpha(agent, price_alpha, model, temp_wealth)
    elif((u_alpha < u_beta) & (u_beta>0) & (price_beta>0) & (price_beta<=temp_wealth)):
        return bid_beta(agent, price_beta, model, temp_wealth)
    elif((u_alpha == u_beta) & (u_alpha>0)):
        return decide_equal_utility(agent, price_alpha, price_beta, model, temp_wealth)
    else: # There is no contract which we can/want to buy so we do not place any bid
        return 0

def place_bids(agent,model):
    temp_wealth = agent.wealth
    remove_from_bid_queues(model, agent.unique_id) # Remove all current bids from the queues
    while(temp_wealth>0): # Every bid should be fullfilable so the wealth should be enough for each bid
        price_alpha = agent.bid_price_alpha(model)
        price_beta = agent.bid_price_beta(model)
        diff = decide_contract(agent, model, price_alpha, price_beta,temp_wealth)
        if diff == 0: # No bid was placed
            return
        temp_wealth = temp_wealth - diff

def remove_from_ask_queues(model, agent_id):
    model.ask_queue_alpha = list(filter(lambda ask: ask.agent.unique_id != agent_id, model.ask_queue_alpha))
    model.ask_queue_beta = list(filter(lambda ask: ask.agent.unique_id != agent_id, model.ask_queue_beta))

def remove_from_bid_queues(model, agent_id):
    model.bid_queue_alpha = list(filter(lambda bid: bid.agent.unique_id != agent_id, model.bid_queue_alpha))
    model.bid_queue_beta = list(filter(lambda bid: bid.agent.unique_id != agent_id, model.bid_queue_beta))

def step_agent(agent, model):
    buy_unit_portfolio(agent,model)
    sell_unit_portfolio(agent,model)
    place_asks(agent,model)  
    place_bids(agent,model) 