from helpers.write_csv import write_to_csv
import random
from mesa import Model
from mesa.time import RandomActivation
from mesa.datacollection import DataCollector
import numpy.ma as ma
import itertools
from model.model_statistics import *
from agents.initialization import *
from agents.agents import step_agent

class MarketModel(Model):
    id_gen = itertools.count(1)
    def __init__(self, num_adaptable, num_fixedmarkup, num_randommarkup, num_strategic, total_wealth, wealth_distribution, belief_distribution, alpha_shape, mean_belief, sigma_belief, csv_name = "NA", width=10, height=10):
        self.uid = next(self.id_gen)
        self.running = True
        self.csv_name = csv_name
        self.step_count = 0
        self.num_agents = num_adaptable + num_fixedmarkup + num_randommarkup + num_strategic 

        self.num_adaptable = num_adaptable
        self.num_fixed_markup = num_fixedmarkup
        self.num_random_markup = num_randommarkup
        self.num_strategic = num_strategic
        self.schedule = RandomActivation(self) # Part of the mesa framework
        
        self.wealth_distribution = wealth_distribution
        self.belief_distribution = belief_distribution
        self.alpha_shape = alpha_shape

        self.mean_belief = mean_belief
        self.sigma_belief = sigma_belief
        self.total_wealth = total_wealth
        self.bid_queue_alpha = []
        self.bid_queue_beta = []
        self.ask_queue_alpha = []
        self.ask_queue_beta = []
        self.unit_portfolios = 0
        
        self.trades_alpha = []
        self.trades_beta = []

        self.check_running()
        # Create agents
        if(self.running == True):
        	init_agents(self)

        self.average_belief = ma.mean([float(a.belief) for a in self.schedule.agents])
        self.median_belief = ma.median([float(a.belief) for a in self.schedule.agents])

        self.datacollector = DataCollector( # Collects the data for the visual representation
            model_reporters={"Price_Alpha": lowest_ask_alpha_price,
                             "Price_Beta": lowest_ask_beta_price,
                             "BidPrice_Alpha": highest_bid_alpha_price,
                             "BidPrice_Beta": highest_bid_beta_price,
                             "LastPrice_Alpha":last_trade_alpha,
                             "LastPrice_Beta": last_trade_beta,
                       		 "Adaptable_agents": get_adaptable,
        					 "Fixedmarkup_agents": get_fixed_markup,
        					 "Randommarkup_agents": get_random_markup,
        					 "Strategic_agents": get_strategic,
        					 "Wealth_distribution": get_wealth_distribution,
        					 "Total_wealth": get_total_wealth,
        					 "Alpha_shape": get_alpha_shape,
        					 "Belief_distribution": get_belief_distribution,
        					 "Mean_belief": get_mean_belief,
        					 "Sigma_belief": get_sigma_belief,
                             "Average_belief": get_average_belief,
                             "Median_belief": get_median_belief,
                             "Trades_Alpha":get_alpha_trades,
                             "Trades_Beta":get_beta_trades,
                             "Beliefs": get_beliefs,
                             "Run": track_run
                            }
                            ) 

    # Check whether the model has reached the end of a simulation
    def check_running(self):
        # The model took 1000 steps, so should end
        if((self.step_count>1000)):
            self.running = False
            self.datacollector.collect(self)
            write_to_csv(self) # Only writes the final data to a csv when it has reached the end state
        # These statements make sure the model does not start unnecessary runs. Since not all the variables
        # are used in a simulation, but they do take on differnt values, certain "default" values are chosen.
        if(self.num_agents!=200):
        	self.running = False
        if((self.wealth_distribution != "Pareto") & (self.alpha_shape != 1.16)):
        	self.running = False
        if((self.belief_distribution != "Normal") & ((self.mean_belief != 0.5) | (self.sigma_belief != 0.125))):
        	self.running = False
        else:
            self.step_count = self.step_count + 1

    # The model will take a random agent and make it perform its actions
    def step(self):
        self.check_running()   # Checks if the model has reached an end state
        self.datacollector.collect(self) # Collects the data per step, used for the visualisation
        if(self.running != False):  
            agent = random.choice(self.schedule.agents) # Picks a random agent
            step_agent(agent, self)